
package scala.tools.nsc
package backend
package icode

import scala.tools.nsc.backend.icode.ICodes


trait ThreeAddressCode { self: ICodes =>
  import global.{Symbol, NoSymbol, Type, Name, Constant};

  abstract case class ThreeAddressStmt {}

  /** `lhs' gets a value assigned from the element of array `arr' at index `index'. */
  case class FROM_ARRAY_ITEM(lhs: Local, arr: Local, index: Local) extends ThreeAddressStmt {
    override def toString(): String = lhs+" = "+arr+"["+index+"]"
  }

  /** `lhs' gets a value assigned from `var' */
  case class FROM_VAR(lhs: Local, rhs: Local) extends ThreeAddressStmt {
    override def toString(): String = lhs+" = "+rhs
  }

  /** `lhs' gets a value assigned from `instance.field' (non-static) or from `field' (static) */
  case class FROM_FIELD(lhs: Local, instance: Local, field: Symbol, isStatic: Boolean) extends ThreeAddressStmt {
    override def toString(): String = {
      val fieldName = (if (isStatic) field.fullName else field.toString())
      val instanceNameDot = (if (isStatic) "" else instance+".")
      lhs+" = "+instanceNameDot+fieldName
    }
  }

  /** `lhs' gets a value assigned from `module' */
  case class FROM_MODULE(lhs: Local, module: Symbol) extends ThreeAddressStmt {
    override def toString(): String = lhs+" = "+module.toString()
  }

  /** `lhs' gets a value assigned from `constant' */
  case class FROM_CONSTANT(lhs: Local, constant: Constant) extends ThreeAddressStmt {
    override def toString(): String = lhs+" = "+constant.toString()
  }


  /** `lhs' gets a value assigned from a primitive function. */
  case class FROM_PRIMITIVE(lhs: Local, primitive: Primitive, args: Array[Local]) extends ThreeAddressStmt {
    override def toString(): String = {
      val expr = (if (args.length == 2) args(0)+" "+primitive.toString()+" "+args(1)
                  else primitive.toString()+args.mkString("(", ",", ")"))
      lhs+" = "+expr
    }
  }

  /** `lhs' gets a value assigned from the result of a method invocation (void is returned as UNIT,
   *   will be desugared later in GenJVM or GenMSIL)  
   *
   * STYLE: dynamic / static(StaticInstance)
   * Stack: ...:ref:arg1:arg2:...:argn
   *    ->: ...:result
   *
   * STYLE: static(StaticClass)
   * Stack: ...:arg1:arg2:...:argn
   *    ->: ...:result
   *
   */
  case class FROM_METHOD_CALL(lhs: Local, method: Symbol, style: opcodes.InvokeStyle, args: Array[Local]) extends ThreeAddressStmt {

    override def toString(): String =
      hostClass.fullName+"."+method.fullName+" ("+args.mkString(",")+") /* InvokeStyle: "+style.toString()+" */";

    def hostClass: Symbol = method.owner;
    /** This is specifically for preserving the target native Array type long
     *  enough that clone() can generate the right call.
     */
    var targetTypeKind: TypeKind = UNIT // the default should never be used, so UNIT should fail fast.
    def setTargetTypeKind(tk: TypeKind) = targetTypeKind = tk

    def consumed = method.tpe.paramTypes.length + (
      style match {
        case opcodes.Dynamic | opcodes.InvokeDynamic => 1
        case opcodes.Static(true) => 1
        case opcodes.Static(false) => 0
        case opcodes.SuperCall(_) => 1
      }
    )

    def consumedTypes = {
      val args = method.tpe.paramTypes map toTypeKind
      style match {
        case opcodes.Dynamic | opcodes.Static(true) => AnyRefReference :: args
        case _ => args
      }
    }

    def produced =
      if(toTypeKind(method.tpe.resultType) == UNIT)
        0
      else if(method.isConstructor)
        0
      else 1

  }

  /**
   * `lhs' gets a value assigned after creating a new instance of a class through the specified constructor
   * Stack: ...:arg1:arg2:...:argn
   *    ->: ...:ref
   */
  case class FROM_NEW(lhs: Local, kind: REFERENCE, method: Symbol, args: Array[Local]) extends ThreeAddressStmt {
    override def toString(): String = lhs+" = new "+ kind + args.mkString("(", ",", ")");
  }

  /**
   *
   * `lhs' gets a value assigned after creating a new array with specified dimensions 
   * Stack: ...:size_1:size_2:..:size_n
   *    ->: ...:arrayref
   */
  case class FROM_NEW_ARRAY(lhs: Local, elem: TypeKind, dimensions: Array[Local]) extends ThreeAddressStmt {
    override def toString(): String = lhs+" = new "+elem.toString()+dimensions.mkString("[",",","]") ;
  }

  /** casts rhs into lhs */
  case class FROM_CAST(lhs: Local, typ: TypeKind, rhs: Local) extends ThreeAddressStmt {
    override def toString(): String = lhs+" = "+rhs+".asInstanceOf["+typ+"]"
  }

  /** An A3_SWITCH stmt compares the index in `input' against those in tags, jumping to the corresponding `label'
   *
   * The tags list contains one entry per label, each entry being in turn
   * a list of ints, any of which will trigger the jump to the corresponding label.
   * labels should contain an extra label, which is the 'default' jump.
   */
  case class A3_SWITCH(input: Local, tags: List[List[Int]], labels: List[BasicBlock]) extends ThreeAddressStmt {
    override def toString(): String ="SWITCH ..." // TODO 
  }

  case class A3_JUMP(whereto: BasicBlock) extends ThreeAddressStmt {
    override def toString(): String ="goto "+whereto.label
  }

  /** An A3_CJUMP stmt compares the two operands using the 'cond' operator,
   * depending on which one of two jumps (to successBlock or to failureBlock) will be taken */
  case class A3_CJUMP(successBlock: BasicBlock, 
                      failureBlock: BasicBlock,
                      cond: TestOp,
                      opA: Local, opB: Local) extends ThreeAddressStmt
  {

    override def toString(): String = (
      "if ("+ opA + TestOpInCondJUMP.strOp(cond) + opB +")"+" goto "+successBlock.label+" else goto "+failureBlock.label
    );

  }

  /** An A3_CJUMP stmt compares the operand and zero using the 'cond' operator */
  case class A3_CZJUMP(successBlock: BasicBlock,
                    failureBlock: BasicBlock,
                    cond: TestOp,
                    opA: Local) extends ThreeAddressStmt {
    override def toString(): String = (
      "if ("+ opA + TestOpInCondJUMP.strOp(cond) + " 0 )"+" goto "+successBlock.label+" else goto "+failureBlock.label
    );

  }


  case class A3_RETURN(result: Local) extends ThreeAddressStmt {
    override def toString(): String ="return "+(if (result != null) result else "")
  }

  case class A3_THROW(result: Local) extends ThreeAddressStmt {
    override def toString(): String ="throw "+result
  }

  case class A3_MONITORENTER(monitor: Local) extends ThreeAddressStmt {
    override def toString(): String ="MONITORENTER "+monitor
  }

  case class A3_MONITOREXIT(monitor: Local) extends ThreeAddressStmt {
    override def toString(): String ="MONITOREXIT "+monitor
  }

  object TestOpInCondJUMP {
    def strOp(cond: TestOp): String = cond match {
        case EQ =>  "=="
        case NE =>  "!="
        case LT =>  "<"
        case GE =>  ">="
        case LE =>  "<="
        case GT =>  ">"
        case _  => throw new RuntimeException("TestOp unknown case")
    }
  }
  /** assigns the value of `rhs' to `instance.field' (non-static) or to `field' (static) */
  case class TO_FIELD(instance: Local, field: Symbol, isStatic: Boolean, rhs: Local) extends ThreeAddressStmt {
    override def toString(): String = {
      val fieldName = (if (isStatic) field.fullName else field.toString())
      val instanceNameDot = (if (isStatic) "" else instance+".")
      instanceNameDot + fieldName+" = "+rhs
    }
  }

}

